"""Register a random.Random-compatible shim for `torch.random` with Hypothesis."""

from typing import Any, Callable

import hypothesis
import hypothesis.internal.entropy
import torch


class TorchRandomWrapper:
    """Shim for `torch.random` to match the random.Random interface.

    Hypothesis's random state manager requires a random.Random-compatible object.
    """

    seed: Callable[..., Any]
    getstate: Callable[[], Any]
    setstate: Callable[..., Any]

    def __init__(self) -> None:
        """Initialize the TorchRandomWrapper."""
        self.seed = torch.random.manual_seed
        self.getstate = torch.random.get_rng_state
        self.setstate = torch.random.set_rng_state


TORCH_RANDOM_WRAPPER = TorchRandomWrapper()


def _register_random_torch_state() -> None:
    """Register the random torch state with Hypothesis."""
    if TORCH_RANDOM_WRAPPER not in hypothesis.internal.entropy.RANDOMS_TO_MANAGE.values():
        hypothesis.register_random(TORCH_RANDOM_WRAPPER)



"""Hypothesis strategies for various Pytorch structures (including tensors and modules).

[Hypothesis](https://hypothesis.readthedocs.io/en/latest/) is a powerful property-based testing library for Python. It
lacks built-in support for Pytorch tensors and modules, so this library provides strategies for generating them.
"""

__version__ = "0.8.7"
import importlib.util

from hypothesis_torch.device import (
    AVAILABLE_CPU_DEVICES,
    AVAILABLE_CUDA_DEVICES,
    AVAILABLE_META_DEVICES,
    AVAILABLE_MPS_DEVICES,
    AVAILABLE_PHYSICAL_DEVICES,
    device_strategy,
)
from hypothesis_torch.dtype import (
    ALL_DTYPES,
    BOOL_DTYPES,
    COMPLEX_DTYPES,
    FLOAT_DTYPES,
    INT_DTYPES,
    SIGNED_INT_DTYPES,
    UNSIGNED_INT_DTYPES,
    dtype_strategy,
)
from hypothesis_torch.layout import layout_strategy
from hypothesis_torch.memory_format import memory_format_strategy
from hypothesis_torch.module import linear_network_strategy, same_shape_activation_strategy
from hypothesis_torch.optim import OptimizerConstructorWithOnlyParameters, optimizer_strategy, optimizer_type_strategy
from hypothesis_torch.register_random_torch_state import TORCH_RANDOM_WRAPPER
from hypothesis_torch.tensor import tensor_strategy

__all__ = [
    "ALL_DTYPES",
    "AVAILABLE_CPU_DEVICES",
    "AVAILABLE_CUDA_DEVICES",
    "AVAILABLE_META_DEVICES",
    "AVAILABLE_MPS_DEVICES",
    "AVAILABLE_PHYSICAL_DEVICES",
    "BOOL_DTYPES",
    "COMPLEX_DTYPES",
    "FLOAT_DTYPES",
    "INT_DTYPES",
    "SIGNED_INT_DTYPES",
    "TORCH_RANDOM_WRAPPER",
    "UNSIGNED_INT_DTYPES",
    "OptimizerConstructorWithOnlyParameters",
    "device_strategy",
    "dtype_strategy",
    "layout_strategy",
    "linear_network_strategy",
    "memory_format_strategy",
    "optimizer_strategy",
    "optimizer_type_strategy",
    "same_shape_activation_strategy",
    "tensor_strategy",
]

if importlib.util.find_spec("transformers") is not None:
    # Import Hugging Face strategies if transformers is installed
    from hypothesis_torch.huggingface import OFFICIALLY_SUPPORTED_TRANSFORMERS, transformer_strategy

    __all__ += ["OFFICIALLY_SUPPORTED_TRANSFORMERS", "transformer_strategy"]



"""Utility functions for inspecting and manipulating callables and classes signatures."""

from __future__ import annotations

import inspect
from typing import Any, Callable, TypeVar

import torch
from hypothesis import strategies as st

T = TypeVar("T")


class ParameterInferAnnotationsFromDefault(inspect.Parameter):
    """A parameter that infers the annotation from the default value if it is not specified.

    This is useful for inferring annotations for parameters that are missing annotations but have default values.
    """

    # mypy does like like the fact that `inspect.Parameter.empty` is a variable pointing to a class.
    @property
    def annotation(self) -> type | inspect.Parameter.empty:  # type: ignore[valid-type]
        """Get the annotation of the parameter, inferring it from the default value if it is not specified.

        Returns:
            The annotation (specified or inferred) of the parameter.

        """
        super_annotation = super().annotation
        if super_annotation is not inspect.Parameter.empty:
            return super_annotation
        if self.default is inspect.Parameter.empty:
            return inspect.Parameter.empty
        return type(self.default)


def infer_signature_annotations(func: Callable) -> inspect.Signature:
    """Get a signature of a callable, inferring missing annotations from default values."""
    sig = inspect.signature(func)
    new_params = []
    for param in sig.parameters.values():
        new_param = ParameterInferAnnotationsFromDefault(
            name=param.name,
            kind=param.kind,
            default=param.default,
        )
        new_params.append(new_param)
    return sig.replace(parameters=new_params)


def get_all_subclasses(cls: type[T]) -> list[type[T]]:
    """Get all subclasses (including descendants) of a class.

    Args:
        cls: The class to get the subclasses of.

    Returns:
        A list of all subclasses of the class.

    """
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))
    return all_subclasses


@st.composite
def signature_to_strategy(
    draw: st.DrawFn,
    constructor: type[T],
    *args: Any,  # noqa: ANN401
    **kwargs: Any,  # noqa: ANN401
) -> T:
    """Strategy for generating instances of a class by drawing values for its constructor.

    Args:
        draw: The draw function provided by `hypothesis`.
        constructor: The class to generate an instance of.
        args: Positional arguments to pass to the constructor. If an argument is a strategy, it will be drawn from.
        kwargs: Keyword arguments to pass to the constructor. If a keyword argument is a strategy, it will be drawn
            from.

    Returns:
        An instance of the class.

    """
    args_drawn = [draw(strategy) for strategy in args]
    kwargs_drawn = {k: draw(strategy) for k, strategy in kwargs.items()}
    return constructor(*args_drawn, **kwargs_drawn)
